dataset_info = dict(
    dataset_name='interhand3d',
    paper_info=dict(
        author='Moon, Gyeongsik and Yu, Shoou-I and Wen, He and '
        'Shiratori, Takaaki and Lee, Kyoung Mu',
        title='InterHand2.6M: A dataset and baseline for 3D '
        'interacting hand pose estimation from a single RGB image',
        container='arXiv',
        year='2020',
        homepage='https://mks0601.github.io/InterHand2.6M/',
    ),
    keypoint_info={
        0:
        dict(
            name='right_thumb4',
            id=0,
            color=[255, 128, 0],
            type='',
            swap='left_thumb4'),
        1:
        dict(
            name='right_thumb3',
            id=1,
            color=[255, 128, 0],
            type='',
            swap='left_thumb3'),
        2:
        dict(
            name='right_thumb2',
            id=2,
            color=[255, 128, 0],
            type='',
            swap='left_thumb2'),
        3:
        dict(
            name='right_thumb1',
            id=3,
            color=[255, 128, 0],
            type='',
            swap='left_thumb1'),
        4:
        dict(
            name='right_forefinger4',
            id=4,
            color=[255, 153, 255],
            type='',
            swap='left_forefinger4'),
        5:
        dict(
            name='right_forefinger3',
            id=5,
            color=[255, 153, 255],
            type='',
            swap='left_forefinger3'),
        6:
        dict(
            name='right_forefinger2',
            id=6,
            color=[255, 153, 255],
            type='',
            swap='left_forefinger2'),
        7:
        dict(
            name='right_forefinger1',
            id=7,
            color=[255, 153, 255],
            type='',
            swap='left_forefinger1'),
        8:
        dict(
            name='right_middle_finger4',
            id=8,
            color=[102, 178, 255],
            type='',
            swap='left_middle_finger4'),
        9:
        dict(
            name='right_middle_finger3',
            id=9,
            color=[102, 178, 255],
            type='',
            swap='left_middle_finger3'),
        10:
        dict(
            name='right_middle_finger2',
            id=10,
            color=[102, 178, 255],
            type='',
            swap='left_middle_finger2'),
        11:
        dict(
            name='right_middle_finger1',
            id=11,
            color=[102, 178, 255],
            type='',
            swap='left_middle_finger1'),
        12:
        dict(
            name='right_ring_finger4',
            id=12,
            color=[255, 51, 51],
            type='',
            swap='left_ring_finger4'),
        13:
        dict(
            name='right_ring_finger3',
            id=13,
            color=[255, 51, 51],
            type='',
            swap='left_ring_finger3'),
        14:
        dict(
            name='right_ring_finger2',
            id=14,
            color=[255, 51, 51],
            type='',
            swap='left_ring_finger2'),
        15:
        dict(
            name='right_ring_finger1',
            id=15,
            color=[255, 51, 51],
            type='',
            swap='left_ring_finger1'),
        16:
        dict(
            name='right_pinky_finger4',
            id=16,
            color=[0, 255, 0],
            type='',
            swap='left_pinky_finger4'),
        17:
        dict(
            name='right_pinky_finger3',
            id=17,
            color=[0, 255, 0],
            type='',
            swap='left_pinky_finger3'),
        18:
        dict(
            name='right_pinky_finger2',
            id=18,
            color=[0, 255, 0],
            type='',
            swap='left_pinky_finger2'),
        19:
        dict(
            name='right_pinky_finger1',
            id=19,
            color=[0, 255, 0],
            type='',
            swap='left_pinky_finger1'),
        20:
        dict(
            name='right_wrist',
            id=20,
            color=[255, 255, 255],
            type='',
            swap='left_wrist'),
        21:
        dict(
            name='left_thumb4',
            id=21,
            color=[255, 128, 0],
            type='',
            swap='right_thumb4'),
        22:
        dict(
            name='left_thumb3',
            id=22,
            color=[255, 128, 0],
            type='',
            swap='right_thumb3'),
        23:
        dict(
            name='left_thumb2',
            id=23,
            color=[255, 128, 0],
            type='',
            swap='right_thumb2'),
        24:
        dict(
            name='left_thumb1',
            id=24,
            color=[255, 128, 0],
            type='',
            swap='right_thumb1'),
        25:
        dict(
            name='left_forefinger4',
            id=25,
            color=[255, 153, 255],
            type='',
            swap='right_forefinger4'),
        26:
        dict(
            name='left_forefinger3',
            id=26,
            color=[255, 153, 255],
            type='',
            swap='right_forefinger3'),
        27:
        dict(
            name='left_forefinger2',
            id=27,
            color=[255, 153, 255],
            type='',
            swap='right_forefinger2'),
        28:
        dict(
            name='left_forefinger1',
            id=28,
            color=[255, 153, 255],
            type='',
            swap='right_forefinger1'),
        29:
        dict(
            name='left_middle_finger4',
            id=29,
            color=[102, 178, 255],
            type='',
            swap='right_middle_finger4'),
        30:
        dict(
            name='left_middle_finger3',
            id=30,
            color=[102, 178, 255],
            type='',
            swap='right_middle_finger3'),
        31:
        dict(
            name='left_middle_finger2',
            id=31,
            color=[102, 178, 255],
            type='',
            swap='right_middle_finger2'),
        32:
        dict(
            name='left_middle_finger1',
            id=32,
            color=[102, 178, 255],
            type='',
            swap='right_middle_finger1'),
        33:
        dict(
            name='left_ring_finger4',
            id=33,
            color=[255, 51, 51],
            type='',
            swap='right_ring_finger4'),
        34:
        dict(
            name='left_ring_finger3',
            id=34,
            color=[255, 51, 51],
            type='',
            swap='right_ring_finger3'),
        35:
        dict(
            name='left_ring_finger2',
            id=35,
            color=[255, 51, 51],
            type='',
            swap='right_ring_finger2'),
        36:
        dict(
            name='left_ring_finger1',
            id=36,
            color=[255, 51, 51],
            type='',
            swap='right_ring_finger1'),
        37:
        dict(
            name='left_pinky_finger4',
            id=37,
            color=[0, 255, 0],
            type='',
            swap='right_pinky_finger4'),
        38:
        dict(
            name='left_pinky_finger3',
            id=38,
            color=[0, 255, 0],
            type='',
            swap='right_pinky_finger3'),
        39:
        dict(
            name='left_pinky_finger2',
            id=39,
            color=[0, 255, 0],
            type='',
            swap='right_pinky_finger2'),
        40:
        dict(
            name='left_pinky_finger1',
            id=40,
            color=[0, 255, 0],
            type='',
            swap='right_pinky_finger1'),
        41:
        dict(
            name='left_wrist',
            id=41,
            color=[255, 255, 255],
            type='',
            swap='right_wrist'),
    },
    skeleton_info={
        0:
        dict(link=('right_wrist', 'right_thumb1'), id=0, color=[255, 128, 0]),
        1:
        dict(link=('right_thumb1', 'right_thumb2'), id=1, color=[255, 128, 0]),
        2:
        dict(link=('right_thumb2', 'right_thumb3'), id=2, color=[255, 128, 0]),
        3:
        dict(link=('right_thumb3', 'right_thumb4'), id=3, color=[255, 128, 0]),
        4:
        dict(
            link=('right_wrist', 'right_forefinger1'),
            id=4,
            color=[255, 153, 255]),
        5:
        dict(
            link=('right_forefinger1', 'right_forefinger2'),
            id=5,
            color=[255, 153, 255]),
        6:
        dict(
            link=('right_forefinger2', 'right_forefinger3'),
            id=6,
            color=[255, 153, 255]),
        7:
        dict(
            link=('right_forefinger3', 'right_forefinger4'),
            id=7,
            color=[255, 153, 255]),
        8:
        dict(
            link=('right_wrist', 'right_middle_finger1'),
            id=8,
            color=[102, 178, 255]),
        9:
        dict(
            link=('right_middle_finger1', 'right_middle_finger2'),
            id=9,
            color=[102, 178, 255]),
        10:
        dict(
            link=('right_middle_finger2', 'right_middle_finger3'),
            id=10,
            color=[102, 178, 255]),
        11:
        dict(
            link=('right_middle_finger3', 'right_middle_finger4'),
            id=11,
            color=[102, 178, 255]),
        12:
        dict(
            link=('right_wrist', 'right_ring_finger1'),
            id=12,
            color=[255, 51, 51]),
        13:
        dict(
            link=('right_ring_finger1', 'right_ring_finger2'),
            id=13,
            color=[255, 51, 51]),
        14:
        dict(
            link=('right_ring_finger2', 'right_ring_finger3'),
            id=14,
            color=[255, 51, 51]),
        15:
        dict(
            link=('right_ring_finger3', 'right_ring_finger4'),
            id=15,
            color=[255, 51, 51]),
        16:
        dict(
            link=('right_wrist', 'right_pinky_finger1'),
            id=16,
            color=[0, 255, 0]),
        17:
        dict(
            link=('right_pinky_finger1', 'right_pinky_finger2'),
            id=17,
            color=[0, 255, 0]),
        18:
        dict(
            link=('right_pinky_finger2', 'right_pinky_finger3'),
            id=18,
            color=[0, 255, 0]),
        19:
        dict(
            link=('right_pinky_finger3', 'right_pinky_finger4'),
            id=19,
            color=[0, 255, 0]),
        20:
        dict(link=('left_wrist', 'left_thumb1'), id=20, color=[255, 128, 0]),
        21:
        dict(link=('left_thumb1', 'left_thumb2'), id=21, color=[255, 128, 0]),
        22:
        dict(link=('left_thumb2', 'left_thumb3'), id=22, color=[255, 128, 0]),
        23:
        dict(link=('left_thumb3', 'left_thumb4'), id=23, color=[255, 128, 0]),
        24:
        dict(
            link=('left_wrist', 'left_forefinger1'),
            id=24,
            color=[255, 153, 255]),
        25:
        dict(
            link=('left_forefinger1', 'left_forefinger2'),
            id=25,
            color=[255, 153, 255]),
        26:
        dict(
            link=('left_forefinger2', 'left_forefinger3'),
            id=26,
            color=[255, 153, 255]),
        27:
        dict(
            link=('left_forefinger3', 'left_forefinger4'),
            id=27,
            color=[255, 153, 255]),
        28:
        dict(
            link=('left_wrist', 'left_middle_finger1'),
            id=28,
            color=[102, 178, 255]),
        29:
        dict(
            link=('left_middle_finger1', 'left_middle_finger2'),
            id=29,
            color=[102, 178, 255]),
        30:
        dict(
            link=('left_middle_finger2', 'left_middle_finger3'),
            id=30,
            color=[102, 178, 255]),
        31:
        dict(
            link=('left_middle_finger3', 'left_middle_finger4'),
            id=31,
            color=[102, 178, 255]),
        32:
        dict(
            link=('left_wrist', 'left_ring_finger1'),
            id=32,
            color=[255, 51, 51]),
        33:
        dict(
            link=('left_ring_finger1', 'left_ring_finger2'),
            id=33,
            color=[255, 51, 51]),
        34:
        dict(
            link=('left_ring_finger2', 'left_ring_finger3'),
            id=34,
            color=[255, 51, 51]),
        35:
        dict(
            link=('left_ring_finger3', 'left_ring_finger4'),
            id=35,
            color=[255, 51, 51]),
        36:
        dict(
            link=('left_wrist', 'left_pinky_finger1'),
            id=36,
            color=[0, 255, 0]),
        37:
        dict(
            link=('left_pinky_finger1', 'left_pinky_finger2'),
            id=37,
            color=[0, 255, 0]),
        38:
        dict(
            link=('left_pinky_finger2', 'left_pinky_finger3'),
            id=38,
            color=[0, 255, 0]),
        39:
        dict(
            link=('left_pinky_finger3', 'left_pinky_finger4'),
            id=39,
            color=[0, 255, 0]),
    },
    joint_weights=[1.] * 42,
    sigmas=[])



import json
from functools import wraps
from logging import log
from paho import mqtt
from paho.mqtt import client as mqtt_client
import random
import platform
import uuid
from alfred.siren.topicgen import get_archives_messages_topic, get_archives_myid_topic, get_archives_rooms_topic, get_chatting_topic, get_events_topic, get_personal_events_topic, get_presence_topic
from alfred.utils.log import logger
from alfred.siren.models import ChatMessage, ContactChat, Invitation, InvitationMessage, InvitationMessageType, PresenceMessage, User, MessageType
import jsons
import time

MQTT_URL = "manaai.cn"
PORT = 1883


class SirenClient:

    def __init__(self, user_acc, user_password, log_level='info') -> None:
        self.user_acc = user_acc
        self.user_password = user_password

        self.client_id = f'{platform.system()}_{uuid.uuid4()}'
        self.user = None
        self.contacts = None
        self.connected = False
        self.client_ready = False

        self._connect()
        self.on_received_msg_func = None
        self.on_received_invitation_func = None
       
        if log_level == 'info':
            logger.remove(handler_id=None)

    def _connect(self):
        self.client = mqtt_client.Client(self.client_id)
        self.client.username_pw_set(self.user_acc, self.user_password)
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.connect(MQTT_URL, PORT)
        logger.info('loop start...')
        self.client.loop_start()
        time.sleep(4)
        while True:
            if not self.connected or not self.client_ready:
                logger.info('still not client ready? wait more 1 seconds.')
                time.sleep(1)
                continue
            else:
                logger.info('Seems connected and client_ready, close loop thread')
                self.client.loop_stop()
                break

    def loop_forever(self):
        self.client.loop_forever()

    def on_connect(self, client, userdata, flags, rc):
        logger.info('connected.. status: ', rc)
        self.subscribe_topics(client)
        if rc == 0:
            self.connected = True

    def on_disconnected(self, client, userdata, flags, rc=0):
        print('disconnected...')

    def subscribe_topics(self, client: mqtt_client):
        client.subscribe(get_archives_rooms_topic(self.client_id))
        client.subscribe(get_archives_messages_topic(self.client_id))
        client.subscribe(get_archives_myid_topic(self.client_id))

    def join_room(self, room_id):
        self.client.subscribe(get_chatting_topic(room_id))
        self.client.subscribe(get_events_topic(room_id))

    def join_contact_presence(self, contact_id):
        self.client.subscribe(get_presence_topic(contact_id))

    def join_my_events(self, myid):
        self.client.subscribe(get_personal_events_topic(myid))

    def on_message(self, client, userdata, msg):
        j = json.loads(msg.payload.decode('utf-8'))
        logger.info(
            '[Msg arrived] topic: {}, payload: {}'.format(msg.topic, j))
        if msg.topic.startswith('archivesrooms/'):
            # join room
            if j != None:
                self.contacts = [jsons.load(i, ContactChat) for i in j]
                for c in self.contacts:
                    self.join_room(c.roomId)
                    self.join_contact_presence(c.id)
                logger.info(f'[Synced all contacts] {len(self.contacts)}')
                self.client_ready = True
        elif msg.topic.startswith("archivesmyid/"):
            # get my id
            self.user = jsons.load(j, User)
            self.join_my_events(self.user.user_addr)
            logger.info(f'Welcome: {self.user.user_nick_name}')
        elif msg.topic.startswith('personalevents/'):
            # invitation auto agree
            invit = jsons.load(j, InvitationMessage)
            logger.info(f'Received invitation: {invit.fromName}')
            self.response_to_invitation(invit.id, invit.fromId)
            self.on_received_invitation(invit)
        elif msg.topic.startswith('messages/'):
            m = jsons.load(j, ChatMessage)
            logger.info(
                f'Received ChatMessage: {m.type} {m.text} {m.fromName} {m.roomId}')
            if m.fromId != self.user.user_addr:
                if self.on_received_msg_func:
                    time.sleep(0.5)
                    self.on_received_msg_func(m)
        elif msg.topic.startswith('presence/'):
            m = jsons.load(j, PresenceMessage)
            logger.info(f'[presence] user {m.fromName} is {m.presenceType}.')
        else:
            logger.info(f'unsupported msg: {j}')

    def response_to_invitation(self, invi_id, sender_id):
        invit = InvitationMessage()
        invit.id = invi_id
        invit.fromId = sender_id
        invit.type = MessageType.EventInvitationResponseAccept
        invit.sendTime = int(round(time.time() * 1000))
        invit.text = "我同意了你的好友请求"
        invit.invitationMessageType = InvitationMessageType.REQUEST_RESPONSE
        j = jsons.dumps(invit)
        if sender_id != None:
            logger.info(f'publish: {sender_id}, {j}')
            self.client.publish(get_personal_events_topic(sender_id), j)

    def on_received_invitation(self, func):
        self.on_received_invitation_func = func

    def on_received_chat_message(self, func):
        self.on_received_msg_func = func

    def publish_txt_msg(self, txt, room_id):
        msg = ChatMessage()
        msg.id = uuid.uuid4()
        msg.roomId = room_id
        msg.fromId = self.user.user_addr
        msg.fromName = self.user.user_nick_name
        msg.text = txt
        msg.sendTime = int(round(time.time() * 1000))
        msg.type = MessageType.ChatText
        j = jsons.dumps(msg)
        t = get_chatting_topic(room_id)
        logger.info(f'publishing: {j}, -> {t}')
        self.client.publish(t, j)

    def publish_img_msg(self, img_url, room_id):
        msg = ChatMessage()
        msg.id = uuid.uuid4()
        msg.roomId = room_id
        msg.fromId = self.user.user_addr
        msg.fromName = self.user.user_nick_name
        msg.text = img_url
        msg.sendTime = int(round(time.time() * 1000))
        msg.type = MessageType.ChatImage
        msg.attachment = img_url
        msg.mime = 'jpg'
        j = jsons.dumps(msg)
        t = get_chatting_topic(room_id)
        logger.info(f'publishing: {j}, -> {t}')
        self.client.publish(t, j)

    """
    For compatible with Uranus old API
    
    """

    def send_txt_msg(self, target_addr, txt):
        contact = [c for c in self.contacts if c.id == target_addr]
        if len(contact) < 1:
            logger.error(
                f'can not send to target since {target_addr} not in contacts.')
        else:
            self.publish_txt_msg(txt, contact[0].roomId)

    def send_img_msg(self, target_addr, img_url):
        contact = [c for c in self.contacts if c.id == target_addr]
        if len(contact) < 1:
            logger.error(
                f'can not send to target since {target_addr} not in contacts.')
        else:
            self.publish_img_msg(img_url, contact[0].roomId)

    def send_msg_to_subscribers(self, txt):
        print(f'----- start broadcast msg to subscribers... {self.client_ready} {self.contacts}')
        if self.client_ready and self.contacts is not None:
            logger.info(
                f'send msg to subscribers, {len(self.contacts)} to go.')
            for c in self.contacts:
                self.publish_txt_msg(txt, c.roomId)
    
    def send_img_msg_to_subscribers(self, img_url):
        print(f'----- start broadcast msg to subscribers... {self.client_ready} {self.contacts}')
        if self.client_ready and self.contacts is not None:
            logger.info(
                f'send msg to subscribers, {len(self.contacts)} to go.')
            for c in self.contacts:
                self.publish_img_msg(img_url, c.roomId)



import numpy as np
from glob import glob
import os
import xml.etree.ElementTree as ET


class KmeansYolo:

    def __init__(self, ann_dir, cluster_number, ann_format='voc'):
        assert ann_format.lower() == 'voc' or ann_format.lower() == 'yolo', 'only support voc or yolo format now.'
        self.cluster_number = cluster_number
        self.ann_dir = ann_dir
        self.ann_format = ann_format.lower()
        print('KMeans clustering on: {} clusters.'.format(cluster_number))

    def iou(self, boxes, clusters):  # 1 box -> k clusters
        n = boxes.shape[0]
        k = self.cluster_number

        box_area = boxes[:, 0] * boxes[:, 1]
        box_area = box_area.repeat(k)
        box_area = np.reshape(box_area, (n, k))

        cluster_area = clusters[:, 0] * clusters[:, 1]
        cluster_area = np.tile(cluster_area, [1, n])
        cluster_area = np.reshape(cluster_area, (n, k))

        box_w_matrix = np.reshape(boxes[:, 0].repeat(k), (n, k))
        cluster_w_matrix = np.reshape(np.tile(clusters[:, 0], (1, n)), (n, k))
        min_w_matrix = np.minimum(cluster_w_matrix, box_w_matrix)

        box_h_matrix = np.reshape(boxes[:, 1].repeat(k), (n, k))
        cluster_h_matrix = np.reshape(np.tile(clusters[:, 1], (1, n)), (n, k))
        min_h_matrix = np.minimum(cluster_h_matrix, box_h_matrix)
        inter_area = np.multiply(min_w_matrix, min_h_matrix)

        result = inter_area / (box_area + cluster_area - inter_area)
        return result

    def avg_iou(self, boxes, clusters):
        accuracy = np.mean([np.max(self.iou(boxes, clusters), axis=1)])
        return accuracy

    def kmeans(self, boxes, k, dist=np.median):
        box_number = boxes.shape[0]
        distances = np.empty((box_number, k))
        last_nearest = np.zeros((box_number,))
        np.random.seed()
        clusters = boxes[np.random.choice(
            box_number, k, replace=False)]  # init k clusters
        while True:

            distances = 1 - self.iou(boxes, clusters)

            current_nearest = np.argmin(distances, axis=1)
            if (last_nearest == current_nearest).all():
                break  # clusters won't change
            for cluster in range(k):
                clusters[cluster] = dist(  # update clusters
                    boxes[current_nearest == cluster], axis=0)

            last_nearest = current_nearest

        return clusters

    def result2txt(self, data):
        f = open("yolo_anchors.txt", 'w')
        row = np.shape(data)[0]
        for i in range(row):
            if i == 0:
                x_y = "%d,%d" % (data[i][0], data[i][1])
            else:
                x_y = ", %d,%d" % (data[i][0], data[i][1])
            f.write(x_y)
        f.close()

    def _load_yolo_boxes(self, ann_files):
        assert isinstance(ann_files, list), 'ann_files must be a list'
        dataSet = []
        for f in ann_files:
            lines = open(f, 'r').readlines()
            infos = [i.strip() for i in lines]
            for line in infos:
                width = int(line.split(" ")[3])
                height = int(line.split(",")[4])
                dataSet.append([width, height])
        return dataSet
    
    def _load_voc_boxes(self, ann_files):
        assert isinstance(ann_files, list), 'ann_files must be a list'
        dataSet = []
        for f in ann_files:
            tree = ET.parse(f)
            root = tree.getroot()
            # size = root.find('size')
            # im_w = int(size.find('width').text)
            # im_h = int(size.find('height').text)

            for obj in root.iter('object'):
                xmlbox = obj.find('bndbox')
               
                w = float(xmlbox.find('xmax').text) - float(xmlbox.find('xmin').text)
                h = float(xmlbox.find('ymax').text) - float(xmlbox.find('ymin').text)
                dataSet.append([w, h])
        return dataSet

    def txt2boxes(self):
        files = None
        dataSet = None
        if self.ann_format == 'yolo':
            files = glob(os.path.join(self.ann_dir, '*.txt'))
            dataSet = self._load_yolo_boxes(files)
            print('[WARN] yolo format anchor calculation **only** support relative anchor (same as YoloV5 format)' +
            'if you want using it in YoloV3, you gonna need x image wh.')
        elif self.ann_format == 'voc':
            files = glob(os.path.join(self.ann_dir, '*.xml'))
            dataSet = self._load_voc_boxes(files)
        print('annotation format: {}, all: {} annos.'.format(self.ann_format, len(files)))
        print('all boxes num: {}'.format(len(dataSet)))
        result = np.array(dataSet)
        return result

    def txt2clusters(self):
        all_boxes = self.txt2boxes()
        print('clustering for anchors....')
        result = self.kmeans(all_boxes, k=self.cluster_number)
        result = result[np.lexsort(result.T[0, None])]
        self.result2txt(result)
        print("K anchors:\n {}".format(result))
        print("Accuracy: {:.2f}%".format(
            self.avg_iou(all_boxes, result) * 100))


if __name__ == "__main__":
    cluster_number = 9
    filename = "2012_train.txt"
    # Row format: image_file_path box1 box2 ... boxN;
    # Box format: x_min,y_min,x_max,y_max,class_id (no space).
    kmeans = KmeansYolo(cluster_number, filename)
    kmeans.txt2clusters()


import platform
import os
import math
import torch
import numpy as np
import cv2
import trimesh


# autopep8: off
os_name = platform.platform().lower()
if 'centos' in os_name or 'windows' in os_name or 'tlinux' in os_name:
    os.environ['PYOPENGL_PLATFORM'] = 'egl'
elif 'debian' in os_name or 'ubuntu' in os_name:
    os.environ['PYOPENGL_PLATFORM'] = 'osmesa'
print(os.environ['PYOPENGL_PLATFORM'])

import pyrender
from pyrender.constants import RenderFlags
from pyrender.constants import DEFAULT_Z_NEAR
# autopep8: on


class WeakPerspectiveCamera(pyrender.Camera):
    def __init__(self,
                 scale,
                 translation,
                 znear=DEFAULT_Z_NEAR,
                 zfar=None,
                 name=None):
        super(WeakPerspectiveCamera, self).__init__(
            znear=znear, zfar=zfar, name=name,
        )
        self.scale = scale
        self.translation = translation

    def get_projection_matrix(self, width=None, height=None):
        P = np.eye(4)
        P[0, 0] = self.scale[0]
        P[1, 1] = self.scale[1]
        P[0, 3] = self.translation[0] * self.scale[0]
        P[1, 3] = -self.translation[1] * self.scale[1]
        P[2, 2] = -1
        return P


class Renderer:
    def __init__(self, smpl_faces, resolution=(224, 224), orig_img=False, wireframe=False, use_gpu=True):
        """
        resolution is: h, w
        """
        self.name = 'pyrender'
        self.resolution = resolution
        self.faces = smpl_faces
        self.orig_img = orig_img
        self.wireframe = wireframe
        self.renderer = pyrender.OffscreenRenderer(
            viewport_width=self.resolution[1],
            viewport_height=self.resolution[0],
            point_size=1.0,
        )
        # set the scene
        self.scene = pyrender.Scene(
            bg_color=[0.0, 0.0, 0.0, 0.0], ambient_light=(0.3, 0.3, 0.3))

        # light = pyrender.PointLight(color=[1.0, 1.0, 1.0], intensity=0.8)
        light = pyrender.DirectionalLight(color=[1.0, 1.0, 1.0], intensity=0.8)

        light_pose = np.eye(4)
        light_pose[:3, 3] = [0, -1, 1]
        self.scene.add(light, pose=light_pose)

        light_pose[:3, 3] = [0, 1, 1]
        self.scene.add(light, pose=light_pose)

        light_pose[:3, 3] = [1, 1, 2]
        self.scene.add(light, pose=light_pose)
        print('render initiated.')

    def render(
        self,
        img,
        verts,
        cam,
        angle=None,
        axis=None,
        mesh_filename=None,
        color=[1.0, 1.0, 0.9],
        rotate=False,
    ):
        mesh = trimesh.Trimesh(vertices=verts, faces=self.faces, process=False)

        Rx = trimesh.transformations.rotation_matrix(
            math.radians(180), [1, 0, 0])
        mesh.apply_transform(Rx)
        
        if rotate:
            rot = trimesh.transformations.rotation_matrix(np.radians(60), [0, 1, 0])
            mesh.apply_transform(rot)

        if angle and axis:
            R = trimesh.transformations.rotation_matrix(
                math.radians(angle), axis)
            mesh.apply_transform(R)

        if mesh_filename is not None:
            mesh.export(mesh_filename)

        sx, sy, tx, ty = cam

        camera = WeakPerspectiveCamera(
            scale=[sx, sy], translation=[tx, ty], zfar=1000.0
        )

        material = pyrender.MetallicRoughnessMaterial(
            metallicFactor=0.0,
            alphaMode="OPAQUE",
            smooth=True,
            wireframe=True,
            roughnessFactor=1.0,
            emissiveFactor=(0.1, 0.1, 0.1),
            baseColorFactor=(color[0], color[1], color[2], 1.0),
        )

        mesh = pyrender.Mesh.from_trimesh(mesh, material=material)
        mesh_node = self.scene.add(mesh, 'mesh')

        camera_pose = np.eye(4)
        cam_node = self.scene.add(camera, pose=camera_pose)

        if self.wireframe:
            render_flags = RenderFlags.RGBA | RenderFlags.ALL_WIREFRAME
        else:
            render_flags = RenderFlags.RGBA
        rgb, _ = self.renderer.render(self.scene, flags=render_flags)

        if rgb.shape[-1] == 4:
            valid_mask = (rgb[:, :, -1] > 0)[:, :, np.newaxis]
            output_img = rgb[:, :, :-1] * valid_mask + (1 - valid_mask) * img
            image = output_img.astype(np.uint8)
        else:
            # rgb could be 3 channel output
            valid_mask = (rgb > 0)
            output_img = rgb * valid_mask + (1 - valid_mask) * img
            image = output_img.astype(np.uint8)

        self.scene.remove_node(mesh_node)
        self.scene.remove_node(cam_node)
        return image



from .pose_dataset_info import DatasetInfo
from .pose_dataset_info import get_dataset_info_by_name
import cv2
import numpy as np
import math
from alfred.vis.image.common import colors


def vis_pose_by_joints(img, poses, joints, color=colors(0), point_color=[255, 255, 255], point_size=3):
    N, num_kps, _ = poses.shape
    for pose in poses:
        for part_id in range(num_kps - 1):
            # if num_kps is 18, then 0 ... 17
            kpt_a_id = joints[part_id][0]
            global_kpt_a_id = pose[kpt_a_id, 0]
            if global_kpt_a_id != -1:
                x_a, y_a = pose[kpt_a_id]
                cv2.circle(img, (int(x_a), int(y_a)), 3, color, -1)
            kpt_b_id = joints[part_id][1]
            global_kpt_b_id = pose[kpt_b_id, 0]
            if global_kpt_b_id != -1:
                x_b, y_b = pose[kpt_b_id]
                cv2.circle(img, (int(x_b), int(y_b)),
                           point_size, point_color, -1)
            if global_kpt_a_id != -1 and global_kpt_b_id != -1:
                cv2.line(img, (int(x_a), int(y_a)),
                         (int(x_b), int(y_b)), color, 2)
    return img


def vis_pose_result(
    img,
    pose_result,
    radius=4,
    thickness=1,
    kpt_score_thr=0.3,
    dataset="TopDownCocoDataset",
    dataset_info=None,
    show=False,
    out_file=None,
):
    """Visualize the detection results on the image.

    Args:
        model (nn.Module): The loaded detector.
        img (str | np.ndarray): Image filename or loaded image.
        pose_result (list[dict]): The results to draw over `img`, [N, 17, 3] for coco body
        radius (int): Radius of circles.
        thickness (int): Thickness of lines.
        kpt_score_thr (float): The threshold to visualize the keypoints.
        skeleton (list[tuple()]): Default None.
        show (bool):  Whether to show the image. Default True.
        out_file (str|None): The filename of the output visualization image.
    """

    # get dataset info
    if dataset_info is None:
        dataset_info = get_dataset_info_by_name(dataset)
        assert dataset_info is not None, '{} dataset not supported built in, you can specific dataset_info manually.'.format(
            dataset)
    else:
        dataset_info = DatasetInfo(dataset_info)

    skeleton = dataset_info.skeleton
    pose_link_color = dataset_info.pose_link_color
    pose_kpt_color = dataset_info.pose_kpt_color
    img = imshow_keypoints(
        img,
        pose_result,
        skeleton=skeleton,
        kpt_score_thr=kpt_score_thr,
        pose_kpt_color=pose_kpt_color,
        pose_link_color=pose_link_color,
        radius=radius,
        thickness=thickness,
        show_keypoint_weight=False,
    )
    if show:
        cv2.imshow("pose result", img)
        cv2.waitKey(0)
    if out_file is not None:
        cv2.imwrite(out_file, img)
    return img


def imshow_keypoints(
    img,
    pose_result,
    skeleton=None,
    kpt_score_thr=0.3,
    pose_kpt_color=None,
    pose_link_color=None,
    radius=4,
    thickness=1,
    show_keypoint_weight=False,
):
    """Draw keypoints and links on an image.

    Args:
            img (str or Tensor): The image to draw poses on. If an image array
                is given, id will be modified in-place.
            pose_result (list[kpts]): The poses to draw. Each element kpts is
                a set of K keypoints as an Kx3 numpy.ndarray, where each
                keypoint is represented as x, y, score.
            kpt_score_thr (float, optional): Minimum score of keypoints
                to be shown. Default: 0.3.
            pose_kpt_color (np.array[Nx3]`): Color of N keypoints. If None,
                the keypoint will not be drawn.
            pose_link_color (np.array[Mx3]): Color of M links. If None, the
                links will not be drawn.
            thickness (int): Thickness of lines.
    """
    img_h, img_w, _ = img.shape

    for kpts in pose_result:
        kpts = np.array(kpts, copy=False)

        # draw each point on image
        if pose_kpt_color is not None:
            if len(pose_kpt_color) != len(kpts):
                c = colors(np.random.randint(50))
                pose_kpt_color = [c for _ in range(len(kpts))]
            # assert len(pose_kpt_color) == len(
                # kpts), 'pose_kpt_color: {} not equal kpts: {}'.format(len(pose_kpt_color), len(kpts))
            for kid, kpt in enumerate(kpts):
                if len(kpt) > 2:
                    x_coord, y_coord, kpt_score = int(
                        kpt[0]), int(kpt[1]), kpt[2]
                    if kpt_score > kpt_score_thr:
                        if show_keypoint_weight:
                            img_copy = img.copy()
                            r, g, b = pose_kpt_color[kid]
                            cv2.circle(
                                img_copy,
                                (int(x_coord), int(y_coord)),
                                radius,
                                (int(r), int(g), int(b)),
                                -1,
                                cv2.LINE_AA,
                            )
                            transparency = max(0, min(1, kpt_score))
                            cv2.addWeighted(
                                img_copy, transparency, img, 1 - transparency, 0, dst=img
                            )
                        else:
                            r, g, b = pose_kpt_color[kid]
                            cv2.circle(
                                img,
                                (int(x_coord), int(y_coord)),
                                radius,
                                (int(r), int(g), int(b)),
                                -1,
                                cv2.LINE_AA,
                            )
                else:
                    x_coord, y_coord = int(kpt[0]), int(kpt[1])
                    if show_keypoint_weight:
                        img_copy = img.copy()
                        r, g, b = pose_kpt_color[kid]
                        cv2.circle(
                            img_copy,
                            (int(x_coord), int(y_coord)),
                            radius,
                            (int(r), int(g), int(b)),
                            -1,
                            cv2.LINE_AA,
                        )
                        transparency = max(0, min(1, kpt_score))
                        cv2.addWeighted(
                            img_copy, transparency, img, 1 - transparency, 0, dst=img
                        )
                    else:
                        r, g, b = pose_kpt_color[kid]
                        cv2.circle(
                            img,
                            (int(x_coord), int(y_coord)),
                            radius,
                            (int(r), int(g), int(b)),
                            -1,
                            cv2.LINE_AA,
                        )

        # draw links
        if skeleton is not None and pose_link_color is not None:
            assert len(pose_link_color) == len(skeleton)
            for sk_id, sk in enumerate(skeleton):
                pos1 = (int(kpts[sk[0], 0]), int(kpts[sk[0], 1]))
                pos2 = (int(kpts[sk[1], 0]), int(kpts[sk[1], 1]))
                if len(kpts[sk[0]]) > 2:
                    if (
                        pos1[0] > 0
                        and pos1[0] < img_w
                        and pos1[1] > 0
                        and pos1[1] < img_h
                        and pos2[0] > 0
                        and pos2[0] < img_w
                        and pos2[1] > 0
                        and pos2[1] < img_h
                        and kpts[sk[0], 2] > kpt_score_thr
                        and kpts[sk[1], 2] > kpt_score_thr
                    ):
                        r, g, b = pose_link_color[sk_id]
                        if show_keypoint_weight:
                            img_copy = img.copy()
                            X = (pos1[0], pos2[0])
                            Y = (pos1[1], pos2[1])
                            mX = np.mean(X)
                            mY = np.mean(Y)
                            length = ((Y[0] - Y[1]) ** 2 +
                                      (X[0] - X[1]) ** 2) ** 0.5
                            angle = math.degrees(
                                math.atan2(Y[0] - Y[1], X[0] - X[1]))
                            stickwidth = 2
                            polygon = cv2.ellipse2Poly(
                                (int(mX), int(mY)),
                                (int(length / 2), int(stickwidth)),
                                int(angle),
                                0,
                                360,
                                1,
                            )
                            cv2.fillConvexPoly(
                                img_copy, polygon, (int(r), int(g), int(b)))
                            transparency = max(
                                0, min(
                                    1, 0.5 * (kpts[sk[0], 2] + kpts[sk[1], 2]))
                            )
                            cv2.addWeighted(
                                img_copy, transparency, img, 1 - transparency, 0, dst=img
                            )
                        else:
                            cv2.line(
                                img,
                                pos1,
                                pos2,
                                (int(r), int(g), int(b)),
                                thickness=thickness,
                                lineType=cv2.LINE_AA,
                            )
                else:
                    if (
                        pos1[0] > 0
                        and pos1[0] < img_w
                        and pos1[1] > 0
                        and pos1[1] < img_h
                        and pos2[0] > 0
                        and pos2[0] < img_w
                        and pos2[1] > 0
                        and pos2[1] < img_h
                    ):
                        r, g, b = pose_link_color[sk_id]
                        if show_keypoint_weight:
                            img_copy = img.copy()
                            X = (pos1[0], pos2[0])
                            Y = (pos1[1], pos2[1])
                            mX = np.mean(X)
                            mY = np.mean(Y)
                            length = ((Y[0] - Y[1]) ** 2 +
                                      (X[0] - X[1]) ** 2) ** 0.5
                            angle = math.degrees(
                                math.atan2(Y[0] - Y[1], X[0] - X[1]))
                            stickwidth = 2
                            polygon = cv2.ellipse2Poly(
                                (int(mX), int(mY)),
                                (int(length / 2), int(stickwidth)),
                                int(angle),
                                0,
                                360,
                                1,
                            )
                            cv2.fillConvexPoly(
                                img_copy, polygon, (int(r), int(g), int(b)))
                            transparency = max(
                                0, min(
                                    1, 0.5 * (kpts[sk[0], 2] + kpts[sk[1], 2]))
                            )
                            cv2.addWeighted(
                                img_copy, transparency, img, 1 - transparency, 0, dst=img
                            )
                        else:
                            cv2.line(
                                img,
                                pos1,
                                pos2,
                                (int(r), int(g), int(b)),
                                thickness=thickness,
                                lineType=cv2.LINE_AA,
                            )

    return img



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from collections import Iterable, defaultdict
from copy import deepcopy
from itertools import chain

import torch
from torch import nn
from torch._utils import _unflatten_dense_tensors
from torch.autograd import Variable
from torch.nn.utils import parameters_to_vector

bn_types = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


def split_bn_bias(layer_groups):
    "Split the layers in `layer_groups` into batchnorm (`bn_types`) and non-batchnorm groups."
    split_groups = []
    for l in layer_groups:
        l1, l2 = [], []
        for c in l.children():
            if isinstance(c, bn_types): l2.append(c)
            else: l1.append(c)
        split_groups += [nn.Sequential(*l1), nn.Sequential(*l2)]
    return split_groups


def get_master(layer_groups, flat_master: bool = False):
    "Return two lists, one for the model parameters in FP16 and one for the master parameters in FP32."
    split_groups = split_bn_bias(layer_groups)
    model_params = [[
        param for param in lg.parameters() if param.requires_grad
    ] for lg in split_groups]
    if flat_master:
        master_params = []
        for lg in model_params:
            if len(lg) != 0:
                mp = parameters_to_vector([param.data.float() for param in lg])
                mp = torch.nn.Parameter(mp, requires_grad=True)
                if mp.grad is None: mp.grad = mp.new(*mp.size())
                master_params.append([mp])
            else:
                master_params.append([])
        return model_params, master_params
    else:
        master_params = [[param.clone().float().detach() for param in lg]
                         for lg in model_params]
        for mp in master_params:
            for param in mp:
                param.requires_grad = True
        return model_params, master_params


def model_g2master_g(model_params, master_params,
                     flat_master: bool = False) -> None:
    "Copy the `model_params` gradients to `master_params` for the optimizer step."
    if flat_master:
        for model_group, master_group in zip(model_params, master_params):
            if len(master_group) != 0:
                master_group[0].grad.data.copy_(
                    parameters_to_vector(
                        [p.grad.data.float() for p in model_group]))
    else:
        for model_group, master_group in zip(model_params, master_params):
            for model, master in zip(model_group, master_group):
                if model.grad is not None:
                    if master.grad is None:
                        master.grad = master.data.new(*master.data.size())
                    master.grad.data.copy_(model.grad.data)
                else:
                    master.grad = None


def master2model(model_params, master_params,
                 flat_master: bool = False) -> None:
    "Copy `master_params` to `model_params`."
    if flat_master:
        for model_group, master_group in zip(model_params, master_params):
            if len(model_group) != 0:
                for model, master in zip(
                        model_group,
                        _unflatten_dense_tensors(master_group[0].data,
                                                 model_group)):
                    model.data.copy_(master)
    else:
        for model_group, master_group in zip(model_params, master_params):
            for model, master in zip(model_group, master_group):
                model.data.copy_(master.data)


def listify(p=None, q=None):
    "Make `p` listy and the same length as `q`."
    if p is None: p = []
    elif isinstance(p, str): p = [p]
    elif not isinstance(p, Iterable): p = [p]
    n = q if type(q) == int else len(p) if q is None else len(q)
    if len(p) == 1: p = p * n
    assert len(p) == n, f'List len mismatch ({len(p)} vs {n})'
    return list(p)


def trainable_params(m: nn.Module):
    "Return list of trainable params in `m`."
    res = filter(lambda p: p.requires_grad, m.parameters())
    return res


def is_tuple(x) -> bool:
    return isinstance(x, tuple)


# copy from fastai.
class OptimWrapper(torch.optim.Optimizer):
    "Basic wrapper around `opt` to simplify hyper-parameters changes."

    def __init__(self, opt, wd, true_wd: bool = False, bn_wd: bool = True):
        # super().__init__(opt.param_groups, dict())
        self.opt, self.true_wd, self.bn_wd = opt, true_wd, bn_wd
        self.opt_keys = list(self.opt.param_groups[0].keys())
        self.opt_keys.remove('params')
        self.read_defaults()
        self.wd = wd

    @classmethod
    def create(cls, opt_func, lr, layer_groups, **kwargs):
        "Create an `optim.Optimizer` from `opt_func` with `lr`. Set lr on `layer_groups`."
        split_groups = split_bn_bias(layer_groups)
        opt = opt_func([{
            'params': trainable_params(l),
            'lr': 0
        } for l in split_groups])
        opt = cls(opt, **kwargs)
        opt.lr, opt.opt_func = listify(lr, layer_groups), opt_func
        return opt

    def new(self, layer_groups):
        "Create a new `OptimWrapper` from `self` with another `layer_groups` but the same hyper-parameters."
        opt_func = getattr(self, 'opt_func', self.opt.__class__)
        split_groups = split_bn_bias(layer_groups)
        opt = opt_func([{
            'params': trainable_params(l),
            'lr': 0
        } for l in split_groups])
        return self.create(
            opt_func,
            self.lr,
            layer_groups,
            wd=self.wd,
            true_wd=self.true_wd,
            bn_wd=self.bn_wd)

    def __repr__(self) -> str:
        return f'OptimWrapper over {repr(self.opt)}.\nTrue weight decay: {self.true_wd}'

    #Pytorch optimizer methods
    def step(self) -> None:
        "Set weight decay and step optimizer."
        # weight decay outside of optimizer step (AdamW)
        if self.true_wd:
            for lr, wd, pg1, pg2 in zip(self._lr, self._wd,
                                        self.opt.param_groups[::2],
                                        self.opt.param_groups[1::2]):
                for p in pg1['params']:
                    p.data.mul_(1 - wd * lr)
                if self.bn_wd:
                    for p in pg2['params']:
                        p.data.mul_(1 - wd * lr)
            self.set_val('weight_decay', listify(0, self._wd))
        self.opt.step()

    def zero_grad(self) -> None:
        "Clear optimizer gradients."
        self.opt.zero_grad()

    #Passthrough to the inner opt.
    def __getstate__(self):
        return self.opt.__getstate__()

    def __setstate__(self, state):
        return self.opt.__setstate__(state)

    def state_dict(self):
        return self.opt.state_dict()

    def load_state_dict(self, state_dict):
        return self.opt.load_state_dict(state_dict)

    def add_param_group(self, param_group):
        return self.opt.add_param_group(param_group)

    def clear(self):
        "Reset the state of the inner optimizer."
        sd = self.state_dict()
        sd['state'] = {}
        self.load_state_dict(sd)

    @property
    def param_groups(self):
        return self.opt.param_groups

    @property
    def defaults(self):
        return self.opt.defaults

    @property
    def state(self):
        return self.opt.state


    #Hyperparameters as properties
    @property
    def lr(self) -> float:
        return self._lr[-1]

    @lr.setter
    def lr(self, val: float) -> None:
        self._lr = self.set_val('lr', listify(val, self._lr))

    @property
    def mom(self) -> float:
        return self._mom[-1]

    @mom.setter
    def mom(self, val: float) -> None:
        if 'momentum' in self.opt_keys:
            self.set_val('momentum', listify(val, self._mom))
        elif 'betas' in self.opt_keys:
            self.set_val('betas', (listify(val, self._mom), self._beta))
        self._mom = listify(val, self._mom)

    @property
    def beta(self) -> float:
        return None if self._beta is None else self._beta[-1]

    @beta.setter
    def beta(self, val: float) -> None:
        "Set beta (or alpha as makes sense for given optimizer)."
        if val is None: return
        if 'betas' in self.opt_keys:
            self.set_val('betas', (self._mom, listify(val, self._beta)))
        elif 'alpha' in self.opt_keys:
            self.set_val('alpha', listify(val, self._beta))
        self._beta = listify(val, self._beta)

    @property
    def wd(self) -> float:
        return self._wd[-1]

    @wd.setter
    def wd(self, val: float) -> None:
        "Set weight decay."
        if not self.true_wd:
            self.set_val(
                'weight_decay', listify(val, self._wd), bn_groups=self.bn_wd)
        self._wd = listify(val, self._wd)

    #Helper functions
    def read_defaults(self) -> None:
        "Read the values inside the optimizer for the hyper-parameters."
        self._beta = None
        if 'lr' in self.opt_keys: self._lr = self.read_val('lr')
        if 'momentum' in self.opt_keys: self._mom = self.read_val('momentum')
        if 'alpha' in self.opt_keys: self._beta = self.read_val('alpha')
        if 'betas' in self.opt_keys:
            self._mom, self._beta = self.read_val('betas')
        if 'weight_decay' in self.opt_keys:
            self._wd = self.read_val('weight_decay')

    def set_val(self, key: str, val, bn_groups: bool = True):
        "Set `val` inside the optimizer dictionary at `key`."
        if is_tuple(val): val = [(v1, v2) for v1, v2 in zip(*val)]
        for v, pg1, pg2 in zip(val, self.opt.param_groups[::2],
                               self.opt.param_groups[1::2]):
            pg1[key] = v
            if bn_groups: pg2[key] = v
        return val

    def read_val(self, key: str):
        "Read a hyperparameter `key` in the optimizer dictionary."
        val = [pg[key] for pg in self.opt.param_groups[::2]]
        if is_tuple(val[0]): val = [o[0] for o in val], [o[1] for o in val]
        return val


class FastAIMixedOptim(OptimWrapper):
    @classmethod
    def create(cls,
               opt_func,
               lr,
               layer_groups,
               model,
               flat_master=False,
               loss_scale=512.0,
               **kwargs):
        "Create an `optim.Optimizer` from `opt_func` with `lr`. Set lr on `layer_groups`."
        opt = OptimWrapper.create(opt_func, lr, layer_groups, **kwargs)
        opt.model_params, opt.master_params = get_master(
            layer_groups, flat_master)
        opt.flat_master = flat_master
        opt.loss_scale = loss_scale
        opt.model = model
        #Changes the optimizer so that the optimization step is done in FP32.
        # opt = self.learn.opt
        mom, wd, beta = opt.mom, opt.wd, opt.beta
        lrs = [lr for lr in opt._lr for _ in range(2)]
        opt_params = [{
            'params': mp,
            'lr': lr
        } for mp, lr in zip(opt.master_params, lrs)]
        opt.opt = opt_func(opt_params)
        opt.mom, opt.wd, opt.beta = mom, wd, beta
        return opt

    def step(self):
        model_g2master_g(self.model_params, self.master_params,
                         self.flat_master)
        for group in self.master_params:
            for param in group:
                param.grad.div_(self.loss_scale)
        super(FastAIMixedOptim, self).step()
        self.model.zero_grad()
        #Update the params from master to model.
        master2model(self.model_params, self.master_params, self.flat_master)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
install alfred into local bin dir.
"""
from setuptools import setup, find_packages
from setuptools import setup, Extension
import io
from os import path

this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


version_file = 'alfred/version.py'


def get_version():
    with open(version_file, 'r') as f:
        exec(compile(f.read(), version_file, 'exec'))
    return locals()['__version__']


setup(name='alfred-py',
      version=get_version(),
      keywords=['deep learning', 'script helper', 'tools'],
      description='Alfred is a DeepLearning utility library.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='GPL-3.0',
      classifiers=[
          # Operation system
          "Operating System :: OS Independent",
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          "Development Status :: 4 - Beta",
          # Indicate who your project is intended for
          "Intended Audience :: Developers",
          # Topics
          "Topic :: Education",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Artificial Intelligence",
          "Topic :: Scientific/Engineering :: Image Recognition",
          # Pick your license as you wish
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
      ],
      packages=[
          'alfred',
          'alfred.dl',
          'alfred.dl.inference',
          'alfred.dl.data',
          'alfred.dl.data.common',
          'alfred.dl.data.meta',
          'alfred.dl.torch',
          'alfred.dl.torch.train',
          'alfred.dl.torch.distribute',
          'alfred.dl.torch.runner',
          'alfred.dl.torch.nn',
          'alfred.dl.torch.nn.modules',
          'alfred.dl.torch.ops',
          'alfred.dl.metrics',
          'alfred.dl.tf',
          'alfred.dl.evaluator',
          'alfred.vis',
          'alfred.modules',
          'alfred.modules.scrap',
          'alfred.modules.text',
          'alfred.modules.vision',
          'alfred.modules.data',
          'alfred.modules.dltool',
          'alfred.modules.cabinet',
          'alfred.modules.cabinet.mdparse',
          'alfred.modules.cabinet.mdparse.formatters',
          'alfred.modules.cabinet.mdparse.transformers',
          'alfred.modules.cabinet.mdparse.transformers.html',
          'alfred.modules.cabinet.mdparse.transformers.md',
          'alfred.modules',
          'alfred.fusion',
          'alfred.vis.image',
          'alfred.vis.image.pose_datasets',
          'alfred.vis.pointcloud',
          'alfred.vis.renders',
          'alfred.vis.mesh3d',
          'alfred.utils',
          'alfred.siren',
          'alfred.protos',
          'alfred.deploy.tensorrt'
      ],
      # package_dir={'alfred': 'alfred'},
      entry_points={
          'console_scripts': [
              'alfred = alfred.alfred:main'
          ]
      },
      include_package_data=True,
      author="Lucas Jin",
      author_email="jinfagang19@163.com",
      url='https://github.com/jinfagang/alfred',
      platforms='any',
      install_requires=['colorama', 'natsort', 'requests', 'regex',
                        'funcy', 'pascal-voc-writer', 'markdown',
                        'future', 'deprecated', 'loguru',
                        'pyquaternion', 'jsons', 'natsort',
                        'loguru', 'pycocotools',
                        'pascal_voc_writer', 'funcy', 'gtrending', 'xdg',
                        'portalocker']
      )




import numpy as np
import cv2
from os.path import join
import os
from .o3d_visconfig import CONFIG
from alfred.utils.log import logger


def calTransformation(v_i, v_j, r, adaptr=False, ratio=10):
    """ from to vertices to T

    Arguments:
        v_i {} -- [description]
        v_j {[type]} -- [description]
    """
    xaxis = np.array([1, 0, 0])
    v = (v_i + v_j)/2
    direc = (v_i - v_j)
    length = np.linalg.norm(direc)
    direc = direc/length
    rotdir = np.cross(xaxis, direc)
    if np.linalg.norm(rotdir) > 1e-3:
        rotdir = rotdir/np.linalg.norm(rotdir)
        rotdir = rotdir * np.arccos(np.dot(direc, xaxis))
        rotmat, _ = cv2.Rodrigues(rotdir)
    else:
        rotmat = np.eye(3)
    # set the minimal radius for the finger and face
    shrink = min(max(length/ratio, 0.005), 0.05)
    eigval = np.array([[length/2/r, 0, 0], [0, shrink, 0], [0, 0, shrink]])
    T = np.eye(4)
    T[:3, :3] = rotmat @ eigval @ rotmat.T
    T[:3, 3] = v
    return T, r, length


class SkelModel:
    def __init__(self, nJoints=None, kintree=None, body_type=None, joint_radius=0.02, res=20, **kwargs) -> None:
        if nJoints is not None:
            self.nJoints = nJoints
            self.kintree = kintree
        else:
            config = CONFIG[body_type]
            self.nJoints = config['nJoints']
            self.kintree = config['kintree']
        self.body_type = body_type
        self.device = 'none'
        cur_dir = os.path.dirname(__file__)
        faces = np.loadtxt(
            join(cur_dir, 'assets', 'sphere_faces_{}.txt'.format(res)), dtype=np.int)
        self.vertices = np.loadtxt(
            join(cur_dir, 'assets', 'sphere_vertices_{}.txt'.format(res)))
        # compose faces
        faces_all = []
        for nj in range(self.nJoints):
            faces_all.append(faces + nj*self.vertices.shape[0])
        for nk in range(len(self.kintree)):
            faces_all.append(faces + self.nJoints *
                             self.vertices.shape[0] + nk*self.vertices.shape[0])
        self.faces = np.vstack(faces_all)
        self.color = None
        self.nVertices = self.vertices.shape[0] * self.nJoints + \
            self.vertices.shape[0] * len(self.kintree)
        self.joint_radius = joint_radius

    def __call__(self, keypoints3d, id=0, return_verts=True, return_tensor=False, **kwargs):
        if len(keypoints3d.shape) == 2:
            keypoints3d = keypoints3d[None]
        if not return_verts:
            return keypoints3d
        if keypoints3d.shape[-1] == 3:  # add confidence
            keypoints3d = np.dstack((keypoints3d, np.ones(
                (keypoints3d.shape[0], keypoints3d.shape[1], 1))))
        r = self.joint_radius
        if keypoints3d.shape[0] > 15 and keypoints3d.shape[0] != self.nJoints:
            logger.warning(f'keypoint3d shape not equal to nJoints: {keypoints3d.shape[0]} vs {self.nJoints}')
        # joints
        min_conf = 0.1
        verts_final = []
        for nper in range(keypoints3d.shape[0]):
            vertices_all = []
            kpts3d = keypoints3d[nper]
            # limb
            closet_joints = []
            for nk, (i, j) in enumerate(self.kintree):
                if kpts3d[i][-1] < min_conf or kpts3d[j][-1] < min_conf:
                    vertices_all.append(self.vertices*0.001)
                    continue
                T, _, length = calTransformation(
                    kpts3d[i, :3], kpts3d[j, :3], r=1)
                if length > 2:  # large than 2 meter
                    vertices_all.append(self.vertices*0.001)
                    continue
                if length < self.joint_radius * 5:
                    closet_joints.append(i)
                    closet_joints.append(j)
                vertices = self.vertices @ T[:3, :3].T + T[:3, 3:].T
                vertices_all.append(vertices)
            for nj in range(self.nJoints):
                if self.body_type in ['bodyhand', 'bodyhandface'] and nj > 25:
                    r_ = r / 2
                else:
                    r_ = r
                if kpts3d[nj, -1] < min_conf:
                    vertices_all.append(self.vertices*0.001)
                    continue
                vertices_all.append(self.vertices*r_ + kpts3d[nj:nj+1, :3])
            vertices = np.vstack(vertices_all)
            verts_final.append(vertices)
        verts_final = np.stack(verts_final)
        return verts_final

    def to(self, none):
        pass

    def merge_params(self, params, share_shape=False):
        keypoints = np.stack([param['keypoints3d'] for param in params])
        return {'keypoints3d': keypoints, 'id': [0]}

    def init_params(self, nFrames):
        return {'keypoints3d': np.zeros((self.nJoints, 4))}


class SMPLSKEL:
    def __init__(self, model_type, gender, body_type) -> None:
        from ..smplmodel import load_model
        config = CONFIG[body_type]
        self.smpl_model = load_model(
            gender, model_type=model_type, skel_type=body_type)
        self.body_model = SkelModel(config['nJoints'], config['kintree'])

    def __call__(self, return_verts=True, **kwargs):
        keypoints3d = self.smpl_model(
            return_verts=False, return_tensor=False, **kwargs)
        if not return_verts:
            return keypoints3d
        else:
            verts = self.body_model(
                return_verts=True, return_tensor=False, keypoints3d=keypoints3d[0])
            return verts

    def init_params(self, nFrames):
        return np.zeros((self.body_model.nJoints, 4))



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

extract VOC image patchs from VOC annotations

"""
"""
read VOC format
annotations and extract image patchs out

"""
import os
import json
import sys
import xml.etree.ElementTree as ET
from alfred.utils.log import logger as logging
import cv2
import argparse


def get(root, name):
    vars = root.findall(name)
    return vars


def get_and_check(root, name, length):
    vars = root.findall(name)
    if len(vars) == 0:
        raise NotImplementedError('Can not find %s in %s.'%(name, root.tag))
    if length > 0 and len(vars) != length:
        raise NotImplementedError('The size of %s is supposed to be %d, but is %d.'%(name, length, len(vars)))
    if length == 1:
        vars = vars[0]
    return vars


def get_filename_as_int(filename):
    try:
        filename = os.path.splitext(filename)[0]
        return int(filename)
    except:
        raise NotImplementedError('Filename %s is supposed to be an integer.'%(filename))


def convert(xml_dir, output_dir, img_dir, xml_list=None):
    assert os.path.join('image dir {} not exist'.format(img_dir))
    os.makedirs(output_dir, exist_ok=True)
    if xml_list:
        list_fp = open(xml_list, 'r')
    else:
        list_fp = os.listdir(xml_dir)
    logging.info('we got all xml files: {}'.format(len(list_fp)))
    json_dict = {"images":[], "type": "instances", "annotations": [],
                 "categories": []}

    i = 0
    for line in list_fp:
        line = line.strip()
        print("Processing %s"%(line))
        xml_f = os.path.join(xml_dir, line)
        tree = ET.parse(xml_f)
        root = tree.getroot()
        path = get(root, 'path')
        if len(path) == 1:
            filename = os.path.basename(path[0].text)
        elif len(path) == 0:
            filename = get_and_check(root, 'filename', 1).text
        else:
            raise NotImplementedError('%d paths found in %s'%(len(path), line))
        # compare filename with xml filename
        if os.path.basename(xml_f).split('.')[0] != filename.split('.')[0]:
            # if not equal, we replace filename with xml file name
            # print('{} != {}'.format(os.path.basename(xml_f).split('.')[0], filename.split('.')[0]))
            filename = os.path.basename(xml_f).split('.')[0] + '.' + filename.split('.')[-1]
            print('revise filename to: {}'.format(filename))
        ## The filename must be a number
        # image_id = get_filename_as_int(filename)
        image_id = i
        size = get_and_check(root, 'size', 1)
        width = int(get_and_check(size, 'width', 1).text)
        height = int(get_and_check(size, 'height', 1).text)
        image = {'file_name': filename, 'height': height, 'width': width,
                 'id':image_id}
        json_dict['images'].append(image)
        ## Cruuently we do not support segmentation
        #  segmented = get_and_check(root, 'segmented', 1).text
        #  assert segmented == '0'
        
        img_root = img_dir
        img_f = os.path.join(img_root, os.path.basename(xml_f).replace('xml', 'jpg'))
        assert img_f, '{} not exist'.format(img_f)
        logging.info('reading image from: {}'.format(img_f))
        ori_img = cv2.imread(img_f)
        bx_id = 0
        for obj in get(root, 'object'):
            category = get_and_check(obj, 'name', 1).text
            bndbox = get_and_check(obj, 'bndbox', 1)
            xmin = float(get_and_check(bndbox, 'xmin', 1).text)
            ymin = float(get_and_check(bndbox, 'ymin', 1).text)
            xmax = float(get_and_check(bndbox, 'xmax', 1).text)
            ymax = float(get_and_check(bndbox, 'ymax', 1).text)
            assert(xmax > xmin)
            assert(ymax > ymin)
            
            # we got an image patch and will save it
            # print('{} - {}, {} - {}'.format(int(xmin), int(xmax), int(ymin), int(ymax)))
            img_patch = ori_img[ int(ymin):int(ymax), int(xmin):int(xmax)]
            # cv2.imshow('rr', img_patch)
            # cv2.waitKey(0)
            # print(category)
            os.makedirs(os.path.join(output_dir, category), exist_ok=True)
            to_save_f = os.path.join(output_dir, category, '{}_{}.jpg'.format(i, bx_id))
            cv2.imwrite(to_save_f, img_patch)
            bx_id += 1
        # image_id plus 1
        i += 1
    logging.info('done.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Extract image patchs from VOC annotations.')
    parser.add_argument('--xml_dir', '-x', type=str, help='xml dir')
    parser.add_argument('--image_dir', '-i', type=str, default='JPEGImages', help='xml dir')
    parser.add_argument('--output_dir', '-o', type=str, default='extracted_out', help='xml dir')
    args = parser.parse_args()

    xml = args.xml_dir
    img = args.image_dir
    output = args.output_dir

    convert(xml, output, img)


#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import datetime
import os
import shutil

def create_folder(prefix, add_time=True, add_str=None, delete=False):
    additional_str = ''
    if delete is True:
        if os.path.exists(prefix):
            shutil.rmtree(prefix)
        os.makedirs(prefix)
    folder = prefix
    if add_time is True:
        # additional_str has a form such as '170903_220351'
        additional_str += datetime.datetime.now().strftime("%y%m%d_%H%M%S")
        if add_str is not None:
            folder += '/' + additional_str + '_' + add_str
        else:
            folder += '/' + additional_str
    if delete is True:
        if os.path.exists(folder):
            shutil.rmtree(folder)
    os.makedirs(folder)
    return folder


#!/usr/bin/python
#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# pip install lxml

import sys
import os
import json
import xml.etree.ElementTree as ET
from alfred.utils.log import logger as logging
import glob
import cv2


START_BOUNDING_BOX_ID = 1
PRE_DEFINE_CATEGORIES = {}
# If necessary, pre-define category and its id
#  PRE_DEFINE_CATEGORIES = {"aeroplane": 1, "bicycle": 2, "bird": 3, "boat": 4,
#  "bottle":5, "bus": 6, "car": 7, "cat": 8, "chair": 9,
#  "cow": 10, "diningtable": 11, "dog": 12, "horse": 13,
#  "motorbike": 14, "person": 15, "pottedplant": 16,
#  "sheep": 17, "sofa": 18, "train": 19, "tvmonitor": 20}


def get(root, name):
    vars = root.findall(name)
    return vars


def get_and_check(root, name, length):
    vars = root.findall(name)
    if len(vars) == 0:
        raise NotImplementedError('Can not find %s in %s.' % (name, root.tag))
    if length > 0 and len(vars) != length:
        raise NotImplementedError(
            'The size of %s is supposed to be %d, but is %d.' % (name, length, len(vars)))
    if length == 1:
        vars = vars[0]
    return vars


def get_filename_as_int(filename):
    try:
        filename = os.path.splitext(filename)[0]
        return int(filename)
    except:
        raise NotImplementedError(
            'Filename %s is supposed to be an integer.' % (filename))


"""
xml_list is optional, we at least need xml_dir and json_file
"""

def get_pure_file_name(file_name):
    file_name = os.path.basename(file_name)
    return os.path.splitext(file_name)[0]
    
def get_file_name_ext(file_name):
    file_name = os.path.basename(file_name)
    return os.path.splitext(file_name)[1]

def convert(xml_dir, img_dir, json_file=None, xml_list=None, index_1=False):
    if index_1:
        print('Annotations save with index start from 1.')
    if xml_list:
        list_fp = open(xml_list, 'r')
    else:
        list_fp = glob.glob(os.path.join(xml_dir, '*.xml'))
    logging.info('we got all xml files: {}'.format(len(list_fp)))
    json_dict = {"images": [], "type": "instances", "annotations": [],
                 "categories": []}
    categories = PRE_DEFINE_CATEGORIES
    bnd_id = START_BOUNDING_BOX_ID

    i = 0
    for line in list_fp:
        line = line.strip()
        if i % 1000 == 0:
            print("Processing %s" % (line))
        # line could be relative path
        xml_f = line
        if not os.path.exists(line):
            xml_f = os.path.join(xml_dir, line)
        tree = ET.parse(xml_f)
        root = tree.getroot()
        path = get(root, 'path')
        if len(path) == 1:
            filename = os.path.basename(path[0].text)
        elif len(path) == 0:
            filename = get_and_check(root, 'filename', 1).text
        else:
            raise NotImplementedError(
                '%d paths found in %s' % (len(path), line))
        # compare filename with xml filename
        if get_pure_file_name(xml_f) != get_pure_file_name(filename):
            # if not equal, we replace filename with xml file name
            # print('{} != {}'.format(os.path.basename(xml_f).split('.')[0], filename.split('.')[0]))

            filename = get_pure_file_name(xml_f) + get_file_name_ext(filename)

            # filename could be wrong sufix
            print('revise filename to: {}'.format(filename))
            if not os.path.exists(os.path.join(os.path.dirname(xml_f), filename)):
                logging.info(
                    'revise filename wrong, try change sufix (but also could be wrong, check your VOC format pls.)')
                filename = get_pure_file_name(filename) + '.jpg'
        
        # filename to jpg name
        img_filename = get_pure_file_name(filename) + '.jpg'
        if not os.path.exists(os.path.join(img_dir, img_filename)):
            img_filename = get_pure_file_name(filename) + '.png'
        # don't support other format

        # The filename must be a number
        # image_id = get_filename_as_int(filename)
        image_id = i
        try:
            size = get_and_check(root, 'size', 1)
            width = int(get_and_check(size, 'width', 1).text)
            height = int(get_and_check(size, 'height', 1).text)
        except Exception as e:
            # if size not found, we infer from image
            logging.info(
                '{} xml format not fully right, force image height, width, this might not be right, but most cases not effect.'.format(xml_f))
            height, width, _ = cv2.imread(os.path.join(img_dir, img_filename)).shape
        
        # check if height or width is normal
        if width <= 0 or height <= 0:
            # we just override width and height stores at xml
            height, width, _ = cv2.imread(os.path.join(img_dir, img_filename)).shape

        image = {'file_name': img_filename, 'height': height, 'width': width,
                 'id': image_id}
        json_dict['images'].append(image)
        # Cruuently we do not support segmentation
        #  segmented = get_and_check(root, 'segmented', 1).text
        #  assert segmented == '0'
        for obj in get(root, 'object'):
            category = get_and_check(obj, 'name', 1).text
            if category not in categories:
                new_id = len(categories)+1 if index_1 else len(categories)
                categories[category] = new_id
            category_id = categories[category]
            bndbox = get_and_check(obj, 'bndbox', 1)
            xmin = float(get_and_check(bndbox, 'xmin', 1).text)
            ymin = float(get_and_check(bndbox, 'ymin', 1).text)
            xmax = float(get_and_check(bndbox, 'xmax', 1).text)
            ymax = float(get_and_check(bndbox, 'ymax', 1).text)
            assert xmax >= xmin, 'xmax: {} xmin: {} check failed.'.format(xmax, xmin)
            assert ymax >= ymin, 'ymax: {} ymin: {} check failed.'.format(ymax, ymin)
            o_width = abs(xmax - xmin)
            o_height = abs(ymax - ymin)
            ann = {'area': o_width*o_height, 'iscrowd': 0, 'image_id':
                   image_id, 'bbox': [xmin, ymin, o_width, o_height],
                   'category_id': category_id, 'id': bnd_id, 'ignore': 0,
                   'segmentation': []}
            json_dict['annotations'].append(ann)
            bnd_id = bnd_id + 1
        # image_id plus 1
        i += 1

    for cate, cid in categories.items():
        cat = {'supercategory': 'none', 'id': cid, 'name': cate}
        json_dict['categories'].append(cat)
    if not json_file:
        json_file = 'annotations_coco.json'
        logging.info(
            'converted coco format will saved into: {}'.format(json_file))
    json_fp = open(json_file, 'w')
    json_str = json.dumps(json_dict)
    json_fp.write(json_str)
    json_fp.close()
    logging.info('done.')


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print('at least 2 auguments are need.')
        print('Usage: %s XML_LIST.txt(optional) XML_DIR OUTPU_JSON.json' %
              (sys.argv[0]))
        exit(1)
    if len(sys.argv) == 3:
        # xml_dir, output_json
        convert(sys.argv[1], sys.argv[2])
    elif len(sys.argv) == 4:
        # xml_list, xml_dir, output_json
        convert(sys.argv[1], sys.argv[2], sys.argv[3])





"""

contains all models used in Siren chatbot API

"""
import json
import jsons
from dataclasses import dataclass


class MessageType:
    ChatText = "ChatText"
    ChatImage = "ChatImage"
    ChatVideo = "ChatVideo"
    ChatAudio = "ChatAudio"
    ChatDocument = "ChatDocument"
    ChatLocation = "ChatLocation"
    ChatContact = "ChatContact"
    EventInvitationRequest = "EventInvitationRequest"
    EventInvitationResponseAccept = "EventInvitationResponseAccept"
    EventInvitationResponseReject = "EventInvitationResponseReject"
    Presence = "Presence"
    ChatMarker = "ChatMarker"
    Typing = "Typing"
    CreateGroup = "CreateGroup"
    RemoveGroup = "RemoveGroup"
    AddUsersToGroup = "AddUsersToGroup"
    RemoveGroupMembers = "RemoveGroupMembers"


class InvitationMessageType:
    REQUEST_RESPONSE = "REQUEST_RESPONSE"
    ERROR = "ERROR"
    INFO = "INFO"


class PresenceType:
    Available = "Available"
    Away = "Away"
    Unavailable = "Unavailable"


class MessageOriginality:
    Original = "Original"
    Reply = "Reply"
    Forward = "Forward"


class PresenceSession:

    def __init__(self) -> None:
        self.id = None
        self.user_id = None
        self.presence = None
        self.last_presence = None


class BaseMessage:

    def __init__(self) -> None:
        self.id = None
        self.type: MessageType = MessageType.ChatText
        self.fromId = None
        self.fromName = None


class PresenceMessage(BaseMessage):

    def __init__(self) -> None:
        super().__init__()

        self.presenceType: PresenceType = PresenceType.Available

class ChatMessage(BaseMessage):

    def __init__(self) -> None:
        super().__init__()

        self.toId = None
        self.toName = None
        self.text = None
        self.attachment = None
        self.thumbnail = None
        self.originalId = None
        self.originalMessage = None
        self.roomId = None
        self.originality: MessageOriginality = MessageOriginality.Original
        self.size = 0
        self.mime = None
        self.sendTime = None
        self.longitude = None
        self.latitude = None


@dataclass
class Invitation:

    def __init__(self) -> None:
        self.id = None
        self.from_id = None
        self.to_id = None
        self.state = None
        self.sent_date = None


class Room:

    def __init__(self) -> None:
        self.id = None
        self.name = None
        self.avatar = None
        self.is_group = None


class RoomMembership:

    def __init__(self) -> None:
        self.room_id = None
        self.user_id = None
        self.role = None


@dataclass
class InvitationMessage:

    def __init__(self) -> None:
        self.id = None
        self.type: MessageType = MessageType.EventInvitationRequest
        self.invitationMessageType: InvitationMessageType = InvitationMessageType.INFO
        self.text = None
        self.sendTime = None
        self.fromId = None
        self.toId = None
        self.fromName = None
        self.fromAvatar = None


@dataclass
class ContactChat:

    def __init__(self) -> None:
        self.firstName = None
        self.lastName = None
        self.id = None
        self.avatar = None
        self.roomId = None
        self.presence: PresenceType = PresenceType.Available
        self.isGroup = None

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)


@dataclass
class User:

    def __init__(self) -> None:
        self.user_addr = None
        self.user_acc = None
        self.user_sign = None
        self.user_nick_name = None



'''
print github trend
'''

# Standard library imports

# Third party imports

from rich.logging import RichHandler
import http.client
from gtrending import fetch_repos
import colorama
from click import secho
import requests
import os
import re
import json
from xdg import xdg_cache_home
from random import randint
import logging
from time import sleep
from datetime import datetime, timedelta
import textwrap
import math
from shutil import get_terminal_size
from rich.align import Align
from rich.console import Console
from rich.console import group as render_group
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from rich.panel import Panel
from rich.columns import Columns
console = Console()

# could be made into config option in the future
CACHED_RESULT_PATH = xdg_cache_home() / "starcli.json"
print(f'temp cached path: {CACHED_RESULT_PATH}')
CACHE_EXPIRATION = 1  # Minutes
API_URL = "https://api.github.com/search/repositories"

date_range_map = {"today": "daily",
                  "this-week": "weekly", "this-month": "monthly"}

status_actions = {
    "retry": "Failed to retrieve data. Retrying in ",
    "invalid": "The server was unable to process the request.",
    "unauthorized": "The server did not accept the credentials. See: https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token",
    "not_found": "The server indicated no data was found.",
    "unsupported": "The request is not supported.",
    "unknown": "An unknown error occurred.",
    "valid": "The request returned successfully, but an unknown exception occurred.",
}

FORMAT = "%(message)s"


httpclient_logger = logging.getLogger("http.client")


def httpclient_logging_debug(level=logging.DEBUG):
    def httpclient_log(*args):
        httpclient_logger.log(level, " ".join(args))

    http.client.print = httpclient_log
    http.client.HTTPConnection.debuglevel = 1


def debug_requests_on():
    """Turn on the logging for requests"""

    logging.basicConfig(
        level=logging.DEBUG,
        format=FORMAT,
        datefmt="[%Y-%m-%d]",
        handlers=[RichHandler()],
    )
    logger = logging.getLogger(__name__)

    from http.client import HTTPConnection

    httpclient_logging_debug()

    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def convert_datetime(date, date_format="%Y-%m-%d"):
    """Safely convert a date string to datetime"""
    try:
        # try to turn the string into a date-time object
        tmp_date = datetime.strptime(date, date_format)
    except ValueError:  # ValueError will be thrown if format is invalid
        secho(
            "Invalid date: " + date + " must be yyyy-mm-dd",
            fg="bright_red",
        )
        return None
    return tmp_date


def get_date(date):
    """Finds the date info in a string"""
    prefix = ""
    if any(i in date[0] for i in [">", "=", "<"]):
        if "=" in date[1]:
            prefix = date[:2]
            date = date.strip(prefix)
        else:
            prefix = date[0]
            date = date.strip(prefix)
    tmp_date = convert_datetime(date)
    if not tmp_date:
        return None
    return prefix + tmp_date.strftime("%Y-%m-%d")


def get_valid_request(url, auth=""):
    """
    Provide a URL to submit a GET request for and handle a connection error.
    """
    while True:
        try:
            session = requests.Session()
            if auth:
                session.auth = (auth.split(":")[0], auth.split(":")[1])
            request = session.get(url)
        except requests.exceptions.ConnectionError:
            secho("Internet connection error...", fg="bright_red")
            return None

        if not request.status_code in (200, 202):
            handling_code = search_error(request.status_code)
            if handling_code == "retry":
                for i in range(15, 0, -1):
                    secho(
                        f"{status_actions[handling_code]} {i} seconds...",
                        fg="bright_yellow",
                    )  # Print and update a timer

                    sleep(1)
            elif handling_code in status_actions:
                secho(status_actions[handling_code], fg="bright_yellow")
                return None
            else:
                secho("An invalid handling code was returned.", fg="bright_red")
                return None
        else:
            break

    return request


def search_error(status_code):
    """
    This returns a directive on how to handle a given HTTP status code.
    """
    int_status_code = int(
        status_code
    )  # Need to make sure the status code is an integer

    http_code_handling = {
        "200": "valid",
        "202": "valid",
        "204": "valid",
        "400": "invalid",
        "401": "unauthorized",
        "403": "retry",
        "404": "not_found",
        "405": "invalid",
        "422": "not_found",
        "500": "invalid",
        "501": "invalid",
    }

    try:
        return http_code_handling[str(int_status_code)]
    except KeyError:
        return "unsupported"


def search(
    language=None,
    created=None,
    pushed=None,
    stars=">=100",
    topics=[],
    user=None,
    debug=False,
    order="desc",
    auth="",
):
    """Returns repositories searched from GitHub API"""
    date_format = "%Y-%m-%d"  # date format in iso format
    if debug:
        debug_requests_on()
        logger = logging.getLogger(__name__)
        logger.debug("Search: created param:" + created)
        logger.debug("Search: order param: " + order)

    day_range = 0 - randint(100, 400)  # random negative from 100 to 400

    if not created:  # if created not provided
        # creation date: the time now minus a random number of days
        # 100 to 400 days - which was stored in day_range
        created_str = ">=" + (datetime.utcnow() + timedelta(days=day_range)).strftime(
            date_format
        )
    else:  # if created is provided
        created_str = get_date(created)
        if not created_str:
            return None

    if not pushed:  # if pushed not provided
        # pushed date: start, is the time now minus a random number of days
        # 100 to 400 days - which was stored in day_range
        pushed_str = ">=" + (datetime.utcnow() + timedelta(days=day_range)).strftime(
            date_format
        )
    else:  # if pushed is provided
        pushed_str = get_date(pushed)
        if not pushed_str:
            return None

    if user:
        query = f"user:{user}+"
    else:
        query = ""

    query += f"stars:{stars}+created:{created_str}"  # construct query
    query += f"+pushed:{pushed_str}"  # add pushed info to query
    # add language to query
    query += f"+language:{language}" if language else ""
    query += f"".join(["+topic:" + i for i in topics])  # add topics to query

    # use query to construct url
    url = f"{API_URL}?q={query}&sort=stars&order={order}"
    if debug:
        logger.debug("Search: url:" + url)  # print the url when debugging
    if debug and auth:
        logger.debug("Auth: on")
    elif debug:
        logger.debug("Auth: off")

    request = get_valid_request(url, auth)
    if request is None:
        return request

    return request.json()["items"]


def search_github_trending(
    language=None, spoken_language=None, order="desc", stars=">=10", date_range=None
):
    """Returns trending repositories from github trending page"""
    if date_range:
        gtrending_repo_list = fetch_repos(
            language, spoken_language, date_range_map[date_range]
        )
    else:
        gtrending_repo_list = fetch_repos(language, spoken_language)
    repositories = []
    for gtrending_repo in gtrending_repo_list:
        repo_dict = convert_repo_dict(gtrending_repo)
        repo_dict["date_range"] = (
            str(repo_dict["date_range"]) + " stars " +
            date_range.replace("-", " ")
            if date_range
            else None
        )
        repo_dict["watchers_count"] = -1  # watchers count not available
        # filter by number of stars
        num = [int(s) for s in re.findall(r"\d+", stars)][0]
        if (
            ("<" in stars and repo_dict["stargazers_count"] < num)
            or ("<=" in stars and repo_dict["stargazers_count"] <= num)
            or (">" in stars and repo_dict["stargazers_count"] > num)
            or (">=" in stars and repo_dict["stargazers_count"] >= num)
        ):
            repositories.append(repo_dict)

    if order == "asc":
        return sorted(repositories, key=lambda repo: repo["stargazers_count"])
    return sorted(repositories, key=lambda repo: repo["stargazers_count"], reverse=True)


def convert_repo_dict(gtrending_repo):
    repo_dict = {}
    repo_dict["full_name"] = gtrending_repo.get("fullname")
    repo_dict["name"] = gtrending_repo.get("name")
    repo_dict["html_url"] = gtrending_repo.get("url")
    repo_dict["stargazers_count"] = gtrending_repo.get("stars", -1)
    repo_dict["language"] = gtrending_repo.get("language")
    # gtrending_repo has key `description` and value is empty string if it's empty
    repo_dict["description"] = (
        gtrending_repo.get("description")
        if gtrending_repo.get("description") != ""
        else None
    )
    repo_dict["date_range"] = gtrending_repo.get("currentPeriodStars")
    return repo_dict


def shorten_count(number):
    """Shortens number"""
    if number < 1000:
        return str(number)

    number = int(number)
    new_number = math.ceil(round(number / 100.0, 1)) * 100

    if new_number % 1000 == 0:
        return str(new_number)[0] + "k"
    if new_number < 1000:
        # returns the same old integer if no changes were made
        return str(number)
    else:
        # returns a new string if the number was shortened
        return str(new_number / 1000.0) + "k"


def get_stats(repo):
    """return formatted string of repo stats"""
    stats = f"{repo['stargazers_count']} ⭐ " if repo["stargazers_count"] != "-1" else ""
    stats += f"{repo['watchers_count']} 👀 " if repo["watchers_count"] != "-1" else ""
    return stats


def list_layout(repos):
    """Displays repositories in list layout using rich"""

    LAYOUT_WIDTH = 80

    @render_group()
    def render_repo(repo):
        """Yields renderables for a single repo."""
        yield Rule(style="bright_yellow")
        yield ""
        # Table with description and stats
        title_table = Table.grid(padding=(0, 1))
        title_table.expand = True
        stats = get_stats(repo)
        title = Text(repo["full_name"], overflow="fold")
        title.stylize(f"yellow link {repo['html_url']}")
        date_range_col = (
            Text(
                ("(+" + repo["date_range"] + ")").replace("stars ", ""),
                style="bold cyan",
            )
            if "date_range" in repo.keys() and repo["date_range"]
            else Text("")
        )
        title_table.add_row(title, Text(
            stats, style="bold blue") + date_range_col)
        title_table.columns[1].no_wrap = True
        title_table.columns[1].justify = "right"
        yield title_table
        yield ""
        lang_table = Table.grid(padding=(0, 1))
        lang_table.expand = True
        language_col = (
            Text(repo["language"], style="bold cyan")
            if repo["language"]
            else Text("unknown language")
        )
        lang_table.add_row(language_col)
        yield lang_table
        yield ""
        # Descripion
        description = repo["description"]
        if description:
            yield Text(description.strip(), style="green")
        else:
            yield "[i green]no description"
        yield ""

    def column(renderable):
        """Constrain width and align to center to create a column."""
        return Align.center(renderable, width=LAYOUT_WIDTH, pad=False)

    for repo in repos:
        console.print(column(render_repo(repo)))
    console.print(column(Rule(style="bright_yellow")))


def table_layout(repos):
    """Displays repositories in a table format using rich"""

    table = Table(leading=1)

    # make the columns
    table.add_column("Name", style="bold cyan")
    table.add_column("Language", style="green")
    table.add_column("Description", style="blue")
    table.add_column("Stats", style="magenta")

    for repo in repos:
        stats = get_stats(repo)
        stats += (
            "\n" + repo["date_range"].replace("stars", "⭐")
            if "date_range" in repo.keys() and repo["date_range"]
            else ""
        )

        if not repo["language"]:  # if language is not provided
            repo["language"] = "None"  # make it a string
        if not repo["description"]:  # same here
            repo["description"] = "None"

        table.add_row(
            repo["full_name"] + '\n' + repo['html_url'],
            repo["language"],  # so that it can work here
            repo["description"],
            stats,
        )
    console.print(table)


def grid_layout(repos):
    """Displays repositories in a grid format using rich"""

    max_desc_len = 90

    panels = []
    for repo in repos:

        stats = get_stats(repo)
        # '\n' added here as it would group both text and new line together
        # hence if date_range isn't present the new line will also not be displayed
        date_range_str = (
            repo["date_range"].replace("stars", "⭐") + "\n"
            if "date_range" in repo.keys() and repo["date_range"]
            else ""
        )

        if not repo["language"]:  # if language is not provided
            repo["language"] = "None"  # make it a string
        if not repo["description"]:
            repo["description"] = "None"

        name = Text(repo["name"], style="bold yellow")
        language = Text(repo["language"], style="magenta")
        description = Text(repo["description"], style="green")
        stats = Text(stats, style="blue")

        # truncate rest of the description if
        # it's more than 90 (max_desc_len) chars
        # using truncate() is better than textwrap
        # because it also takes care of asian characters
        description.truncate(max_desc_len, overflow="ellipsis")

        repo_summary = Text.assemble(
            name,
            "\n",
            stats,
            "\n",
            date_range_str,
            language,
            "\n",
            description,
        )
        panels.append(Panel(repo_summary, expand=True))

    console.print((Columns(panels, width=30, expand=True)))


def print_results(*args, page=False, layout=""):
    """Use a specified layout to print or page the fetched results"""
    if page:
        with console.pager():
            print_layout(layout=layout, *args)
    else:
        print_layout(
            layout=layout,
            *args,
        )


def print_layout(*args, layout="list"):
    if layout == "table":
        table_layout(*args)
    elif layout == "grid":
        grid_layout(*args)
    else:
        list_layout(*args)
    return


def get_github_trending(
    lang='',
    spoken_language='',
    created='',
    topic=[],
    pushed='',
    layout='table',
    stars='>=100',
    limit_results=50,
    order='desc',
    long_stats=True,
    # date_range='today',
    date_range='this-week',
    user='',
    debug=False,
    auth="",
    pager=False,
):
    """Find trending repos on GitHub"""
    if debug:
        import logging

        debug_requests_on()

    tmp_repos = None
    options_key = "{lang}_{spoken_language}_{created}_{topic}_{pushed}_{stars}_{order}_{date_range}_{user}".format(
        lang=lang,
        spoken_language=spoken_language,
        created=created,
        topic=topic,
        pushed=pushed,
        stars=stars,
        order=order,
        date_range=date_range,
        user=user,
    )

    if os.path.exists(CACHED_RESULT_PATH):
        with open(CACHED_RESULT_PATH, "r") as f:
            json_file = json.load(f)
            result = json_file.get(options_key)
            if result:
                t = result[-1].get("time")
                time = datetime.strptime(t, "%Y-%m-%d %H:%M:%S.%f")
                diff = datetime.now() - time
                if diff < timedelta(minutes=CACHE_EXPIRATION):
                    if debug:
                        logger = logging.getLogger(__name__)
                        logger.debug("Fetching results from cache")

                    tmp_repos = result

    if not tmp_repos:  # If cache expired or results not yet cached
        if auth and not re.search(".:.", auth):  # Check authentication format
            print(
                f"Invalid authentication format: {auth} must be 'username:token'",
            )
            print(
                "Use --help or see: https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token",
            )
            auth = None

        if (
            not spoken_language and not date_range
        ):  # if filtering by spoken language and date range not required
            tmp_repos = search(
                lang, created, pushed, stars, topic, user, debug, order, auth
            )
        else:
            tmp_repos = search_github_trending(
                lang, spoken_language, order, stars, date_range
            )

        if not tmp_repos:  # if search() returned None
            return
        else:  # Cache results
            tmp_repos.append({"time": str(datetime.now())})
            with open(CACHED_RESULT_PATH, "a+") as f:
                if os.path.getsize(CACHED_RESULT_PATH) == 0:  # file is empty
                    result_dict = {options_key: tmp_repos}
                    f.write(json.dumps(result_dict, indent=4))
                else:  # file is not empty
                    f.seek(0)
                    result_dict = json.load(f)
                    result_dict[options_key] = tmp_repos
                    f.truncate(0)
                    f.write(json.dumps(result_dict, indent=4))
    tmp_repos = [r for r in tmp_repos if 'name' in r.keys()
                 and 'full_name' in r.keys()]
    repos = tmp_repos[0:limit_results]

    if not long_stats:  # shorten the stat counts when not --long-stats
        for repo in repos:
            repo["stargazers_count"] = shorten_count(repo["stargazers_count"])
            repo["watchers_count"] = shorten_count(repo["watchers_count"])
            if "date_range" in repo.keys() and repo["date_range"]:
                num_stars = repo["date_range"].split()[0]
                repo["date_range"] = repo["date_range"].replace(
                    num_stars, str(shorten_count(
                        int(num_stars.replace(",", ""))))
                )

    print_results(repos, page=pager, layout=layout)




import importlib
from yacs.config import CfgNode as CN
import argparse


class Config:
    @classmethod
    def load_from_args(cls):
        parser = argparse.ArgumentParser()
        parser.add_argument('--cfg', type=str, default='config/base.yml')
        parser.add_argument("opts", default=None, nargs=argparse.REMAINDER)
        args = parser.parse_args()
        return cls.load(filename=args.cfg, opts=args.opts)

    @classmethod
    def load(cls, filename=None, opts=[]) -> CN:
        cfg = CN()
        cfg = cls.init(cfg)
        if filename is not None:
            cfg.merge_from_file(filename)
        if len(opts) > 0:
            cfg.merge_from_list(opts)
        cls.parse(cfg)
        cls.print(cfg)
        return cfg

    @staticmethod
    def init(cfg):
        return cfg

    @staticmethod
    def parse(cfg):
        pass

    @staticmethod
    def print(cfg):
        print('[Info] --------------')
        print('[Info] Configuration:')
        print('[Info] --------------')
        print(cfg)


def load_object(module_name, module_args):
    module_path = '.'.join(module_name.split('.')[:-1])
    module = importlib.import_module(module_path)
    name = module_name.split('.')[-1]
    obj = getattr(module, name)(**module_args)
    return obj



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

Fusion for nuScenes dataset

"""
import numpy as np
from pyquaternion import Quaternion


def project_cam_coords_to_pixel(pts3d_cam, intrinsic):
    """
    project pts3d_cam on image
    such as:

    [[-1.9, 0.12, 37]] -> [[222, 345]]
    """
    intrinsic = np.array(intrinsic)
    pts3d_cam = np.array(pts3d_cam)
    if pts3d_cam.shape[0] != 3:
        # if not hstack, try to transpose it
        pts3d_cam = pts3d_cam.T
    assert pts3d_cam.shape[0] == 3, 'pts3d_cam must be 3 rows.'
    assert intrinsic.shape == (3, 3), 'intrinsic should be 3x3.'
    a = np.dot(intrinsic, pts3d_cam)
    a = a/a[-1, :]
    return a.T[:, :2]

# def compute_3d_box_cam_coords_nuscenes(xyz, lwh, quarternion):
#     """
#         nuScenes camera coordinates using -y as up
#         using quarternion represents rotation of box
#     """
#     # we get rotation_y from quarternion first
#     quarternion = Quaternion(axis=quarternion[1:], angle=quarternion[0])
#     trans_m = transform_matrix(xyz, quarternion)
#     l, w, h = lwh[0], lwh[1], lwh[2]
#     x_corners = [l / 2, l / 2, -l / 2, -l / 2, l / 2, l / 2, -l / 2, -l / 2]
#     y_corners = [0, 0, 0, 0, -h, -h, -h, -h]
#     # y_corners = [0, 0, 0, 0, h, h, h, h]
#     z_corners = [w / 2, -w / 2, -w / 2, w / 2, w / 2, -w / 2, -w / 2, w / 2]

#     corners = np.array([x_corners, y_corners, z_corners], dtype=np.float32)
#     corners4d = np.ones((corners.shape[0]+1, corners.shape[1]))
#     corners4d[:-1, :] = corners
#     print('transformation matrix: ', trans_m)
#     print('corners4d: ', corners4d)
#     corners4d_trans = np.dot(trans_m, corners4d)
#     print('corners4d_trans: ', corners4d_trans)
#     return corners4d_trans


def compute_3d_box_cam_coords_nuscenes(xyz, lwh, quarternion):
    """
        nuScenes camera coordinates using -y as up
        using quarternion represents rotation of box

        only calculate rotation_y?: arcsin(2(wy-zx))
    """
    # we get rotation_y from quarternion first
    if isinstance(quarternion, Quaternion):
        l, w, h = lwh[0], lwh[1], lwh[2]
        x_corners = [l / 2, l / 2, -l / 2, -l / 2, l / 2, l / 2, -l / 2, -l / 2]
        # y_corners = [0, 0, 0, 0, -h, -h, -h, -h]
        y_corners = [-h/2, -h/2, -h/2, -h/2, h/2, h/2, h/2, h/2]
        z_corners = [w / 2, -w / 2, -w / 2, w / 2, w / 2, -w / 2, -w / 2, w / 2]
        corners = np.array([x_corners, y_corners, z_corners], dtype=np.float32)

        rotation_y = np.pi/2 - quarternion.radians
        c, s = np.cos(rotation_y), np.sin(rotation_y)
        R = np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]], dtype=np.float32)

        corners_trans = np.dot(R, corners)
        corners_trans += [[i] for i in xyz]
        return corners_trans
    else:
        raise ValueError('quarternion must be a Quaternion object, make sure '\
            'you using pyquarternion.')


def load_pc_from_file(pc_f):
    # nuScenes lidar is 5 digits one line (last one the ring index)
    return np.fromfile(pc_f, dtype=np.float32, count=-1).reshape([-1, 5])


#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import functools
import inspect
import sys
from collections import OrderedDict

import numba
import numpy as np
import torch


def get_pos_to_kw_map(func):
    pos_to_kw = {}
    fsig = inspect.signature(func)
    pos = 0
    for name, info in fsig.parameters.items():
        if info.kind is info.POSITIONAL_OR_KEYWORD:
            pos_to_kw[pos] = name
        pos += 1
    return pos_to_kw


def get_kw_to_default_map(func):
    kw_to_default = {}
    fsig = inspect.signature(func)
    for name, info in fsig.parameters.items():
        if info.kind is info.POSITIONAL_OR_KEYWORD:
            if info.default is not info.empty:
                kw_to_default[name] = info.default
    return kw_to_default


def change_default_args(**kwargs):
    def layer_wrapper(layer_class):
        class DefaultArgLayer(layer_class):
            def __init__(self, *args, **kw):
                pos_to_kw = get_pos_to_kw_map(layer_class.__init__)
                kw_to_pos = {kw: pos for pos, kw in pos_to_kw.items()}
                for key, val in kwargs.items():
                    if key not in kw and kw_to_pos[key] > len(args):
                        kw[key] = val
                super().__init__(*args, **kw)

        return DefaultArgLayer

    return layer_wrapper

def torch_to_np_dtype(ttype):
    type_map = {
        torch.float16: np.dtype(np.float16),
        torch.float32: np.dtype(np.float32),
        torch.float16: np.dtype(np.float64),
        torch.int32: np.dtype(np.int32),
        torch.int64: np.dtype(np.int64),
        torch.uint8: np.dtype(np.uint8),
    }
    return type_map[ttype]


def check_tensor_equal(t_a, t_b, epsilon=1e-5):
    res = torch.isclose(t_a, t_b, epsilon)
    res2 = torch.all(res)
    res2 = res2.detach().cpu().numpy()
    if res2:
        return res2, None
    else:
        return res2, res


def torch_load_state_dict_without_module(ckp_file):
    """
    this function using for load a model without module
    """
    checkpoint = torch.load(ckp_file)
    state_dict =checkpoint['state_dict']

    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        if 'module.' in k:
            k = k[7:] # remove 'module.' of dataparallel
        new_state_dict[k]=v
    return new_state_dict